@@ -176,6 +176,8 @@ static SCM scm_unbound_p (SCM obj);
 static SCM scm_assert_bound (SCM value, SCM obj);
 static SCM scm_at_assert_bound_ref (SCM obj, SCM index);
 static SCM scm_sys_goops_loaded (void);
+static SCM scm_make_extended_class_from_symbol (SCM type_name_sym, 
+						int applicablep);
 
 /* This function is used for efficient type dispatch.  */
 SCM_DEFINE (scm_class_of, "class-of", 1, 0, 0,
@@ -241,6 +243,7 @@ SCM_DEFINE (scm_class_of, "class-of", 1, 0, 0,
 	  else
 	    return scm_class_procedure;
 	case scm_tc7_gsubr:
+	case scm_tc7_program:
 	  return scm_class_procedure;
 	case scm_tc7_pws:
 	  return scm_class_procedure_with_setter;
@@ -280,9 +283,9 @@ SCM_DEFINE (scm_class_of, "class-of", 1, 0, 0,
 	      else
 		{
 		  SCM name = SCM_STRUCT_TABLE_NAME (SCM_CDR (handle));
-		  SCM class = scm_make_extended_class (scm_is_true (name)
-						       ? scm_i_symbol_chars (name)
-						       : 0,
+		  SCM class = scm_make_extended_class_from_symbol (scm_is_true (name)
+						       ? name
+						       : scm_nullstr,
 						       SCM_I_OPERATORP (x));
 		  SCM_SET_STRUCT_TABLE_CLASS (SCM_CDR (handle), class);
 		  return class;
@@ -746,9 +749,8 @@ SCM_DEFINE (scm_sys_inherit_magic_x, "%inherit-magic!", 2, 0, 0,
       ls = SCM_CDR (ls);
     }
   flags &= SCM_CLASSF_INHERIT;
-  if (flags & SCM_CLASSF_ENTITY)
-    SCM_SET_CLASS_DESTRUCTOR (class, scm_struct_free_entity);
-  else
+
+  if (! (flags & SCM_CLASSF_ENTITY))
     {
       long n = SCM_I_INUM (SCM_SLOT (class, scm_si_nfields));
 #if 0
@@ -765,7 +767,6 @@ SCM_DEFINE (scm_sys_inherit_magic_x, "%inherit-magic!", 2, 0, 0,
 	{
 	  /* NOTE: The following depends on scm_struct_i_size. */
 	  flags |= SCM_STRUCTF_LIGHT + n * sizeof (SCM); /* use light representation */
-	  SCM_SET_CLASS_DESTRUCTOR (class, scm_struct_free_light);
 	}
     }
   SCM_SET_CLASS_FLAGS (class, flags);
@@ -1256,10 +1257,7 @@ SCM_DEFINE (scm_sys_fast_slot_set_x, "%fast-slot-set!", 3, 0, 0,
 #undef FUNC_NAME
 
 
-SCM_SYNTAX (s_atslot_ref, "@slot-ref", scm_i_makbimacro, scm_m_atslot_ref);
-SCM_SYNTAX (s_atslot_set_x, "@slot-set!", scm_i_makbimacro, scm_m_atslot_set_x);
-
-
+
 /** Utilities **/
 
 /* In the future, this function will return the effective slot
@@ -1528,11 +1526,11 @@ wrap_init (SCM class, SCM *m, long n)
 {
   long i;
   scm_t_bits slayout = SCM_STRUCT_DATA (class)[scm_vtable_index_layout];
-  const char *layout = scm_i_symbol_chars (SCM_PACK (slayout));
+  SCM layout = SCM_PACK (slayout);
 
   /* Set all SCM-holding slots to unbound */
   for (i = 0; i < n; i++)
-    if (layout[i*2] == 'p')
+    if (scm_i_symbol_ref (layout, i*2) == 'p')
       m[i] = SCM_GOOPS_UNBOUND;
     else
       m[i] = 0;
@@ -1890,6 +1888,11 @@ typedef struct t_extension {
   SCM extension;
 } t_extension;
 
+
+/* Hint for `scm_gc_malloc ()' et al. when allocating `t_extension'
+   objects.  */
+static const char extension_gc_hint[] = "GOOPS extension";
+
 static t_extension *extensions = 0;
 
 SCM_VARIABLE (scm_var_make_extended_generic, "make-extended-generic");
@@ -1910,7 +1913,8 @@ scm_c_extend_primitive_generic (SCM extended, SCM extension)
     }
   else
     {
-      t_extension *e = scm_malloc (sizeof (t_extension));
+      t_extension *e = scm_gc_malloc (sizeof (t_extension),
+				      extension_gc_hint);
       t_extension **loc = &extensions;
       /* Make sure that extensions are placed before their own
        * extensions in the extensions list.  O(N^2) algorithm, but
@@ -1933,7 +1937,6 @@ setup_extended_primitive_generics ()
       t_extension *e = extensions;
       scm_c_extend_primitive_generic (e->extended, e->extension);
       extensions = e->next;
-      free (e);
     }
 }
 
@@ -2682,6 +2685,34 @@ make_class_from_template (char const *template, char const *type_name, SCM super
   return class;
 }
 
+static SCM
+make_class_from_symbol (SCM type_name_sym, SCM supers, int applicablep)
+{
+  SCM class, name;
+  if (type_name_sym != SCM_BOOL_F)
+    {
+      name = scm_string_append (scm_list_3 (scm_from_locale_string ("<"),
+					    scm_symbol_to_string (type_name_sym),
+					    scm_from_locale_string (">")));
+      name = scm_string_to_symbol (name);
+    }
+  else
+    name = SCM_GOOPS_UNBOUND;
+
+  class = scm_permanent_object (scm_basic_make_class (applicablep
+						      ? scm_class_procedure_class
+						      : scm_class_class,
+						      name,
+						      supers,
+						      SCM_EOL));
+
+  /* Only define name if doesn't already exist. */
+  if (!SCM_GOOPS_UNBOUNDP (name)
+      && scm_is_false (scm_module_variable (scm_module_goops, name)))
+    DEFVAR (name, class);
+  return class;
+}
+
 SCM
 scm_make_extended_class (char const *type_name, int applicablep)
 {
@@ -2693,6 +2724,16 @@ scm_make_extended_class (char const *type_name, int applicablep)
 				   applicablep);
 }
 
+static SCM
+scm_make_extended_class_from_symbol (SCM type_name_sym, int applicablep)
+{
+  return make_class_from_symbol (type_name_sym,
+				 scm_list_1 (applicablep
+					     ? scm_class_applicable
+					     : scm_class_top),
+				 applicablep);
+}
+
 void
 scm_i_inherit_applicable (SCM c)
 {
@@ -2785,11 +2826,16 @@ static SCM
 make_struct_class (void *closure SCM_UNUSED,
 		   SCM vtable, SCM data, SCM prev SCM_UNUSED)
 {
-  if (scm_is_true (SCM_STRUCT_TABLE_NAME (data)))
-    SCM_SET_STRUCT_TABLE_CLASS (data,
-				scm_make_extended_class
-				(scm_i_symbol_chars (SCM_STRUCT_TABLE_NAME (data)),
-				 SCM_CLASS_FLAGS (vtable) & SCM_CLASSF_OPERATOR));
+  SCM sym = SCM_STRUCT_TABLE_NAME (data);
+  if (scm_is_true (sym))
+    {
+      int applicablep = SCM_CLASS_FLAGS (vtable) & SCM_CLASSF_OPERATOR;
+
+      SCM_SET_STRUCT_TABLE_CLASS (data, 
+				  scm_make_extended_class_from_symbol (sym, applicablep));
+    }
+
+  scm_remember_upto_here_2 (data, vtable);
   return SCM_UNSPECIFIED;
 }
 
@@ -2853,7 +2899,6 @@ scm_make_class (SCM meta, char *s_name, SCM supers, size_t size,
     }
   else if (size > 0)
     {
-      SCM_SET_CLASS_DESTRUCTOR (class, scm_struct_free_light);
       SCM_SET_CLASS_INSTANCE_SIZE (class, size);
     }
 
