@@ -1,4 +1,4 @@
-/* Copyright (C) 1995,1996,1997,1998,2000,2001,2002,2003, 2006, 2008 Free Software Foundation, Inc.
+/* Copyright (C) 1995,1996,1997,1998,2000,2001,2002,2003, 2006, 2008, 2009 Free Software Foundation, Inc.
  * 
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -97,16 +97,6 @@ macro_print (SCM macro, SCM port, scm_print_state *pstate)
   return 1;
 }
 
-static SCM
-macro_mark (SCM macro)
-{
-  if (SCM_MACRO_IS_EXTENDED (macro))
-    { scm_gc_mark (SCM_SMOB_OBJECT_2 (macro));
-      scm_gc_mark (SCM_SMOB_OBJECT_3 (macro));
-    }
-  return SCM_SMOB_OBJECT (macro);
-}
-
 static SCM
 makmac (SCM code, scm_t_bits flags)
 {
@@ -339,7 +329,6 @@ void
 scm_init_macros ()
 {
   scm_tc16_macro = scm_make_smob_type ("macro", 0);
-  scm_set_smob_mark (scm_tc16_macro, macro_mark);
   scm_set_smob_print (scm_tc16_macro, macro_print);
 #include "libguile/macros.x"
 }
