@@ -36,13 +36,11 @@
 #define scm_keyword_obarray scm_sys_protects[4]
 #define scm_stand_in_procs scm_sys_protects[5]
 #define scm_object_whash scm_sys_protects[6]
-#define scm_permobjs scm_sys_protects[7]
-#define scm_asyncs scm_sys_protects[8]
-#define scm_protects scm_sys_protects[9]
-#define scm_properties_whash scm_sys_protects[10]
-#define scm_gc_registered_roots scm_sys_protects[11]
-#define scm_source_whash scm_sys_protects[12]
-#define SCM_NUM_PROTECTS 13
+#define scm_asyncs scm_sys_protects[7]
+#define scm_protects scm_sys_protects[8]
+#define scm_properties_whash scm_sys_protects[9]
+#define scm_source_whash scm_sys_protects[10]
+#define SCM_NUM_PROTECTS 11
 
 SCM_API SCM scm_sys_protects[];
 
