@@ -26,6 +26,9 @@
 #include "libguile/__scm.h"
 #include "libguile/print.h"
 
+#include "libguile/bdw-gc.h"
+
+
 
 /* This is the internal representation of a smob type */
 
@@ -45,40 +48,77 @@ typedef struct scm_smob_descriptor
   int gsubr_type; /* Used in procprop.c */
 } scm_smob_descriptor;
 
+
 
+SCM_API SCM scm_i_new_smob_with_mark_proc (scm_t_bits tc,
+					   scm_t_bits, scm_t_bits, scm_t_bits);
 
-#define SCM_NEWSMOB(z, tc, data) \
-do { \
-  z = scm_cell ((tc), (scm_t_bits) (data)); \
-} while (0)
 
-#define SCM_RETURN_NEWSMOB(tc, data) \
-  do { SCM __SCM_smob_answer; \
-       SCM_NEWSMOB (__SCM_smob_answer, (tc), (data)); \
-       return __SCM_smob_answer; \
+
+#define SCM_NEWSMOB(z, tc, data)					  \
+do									  \
+  {									  \
+    register scm_t_bits _smobnum = SCM_TC2SMOBNUM (tc);			  \
+    z = (scm_smobs[_smobnum].mark					  \
+	 ? scm_i_new_smob_with_mark_proc ((tc), (scm_t_bits)(data),	  \
+					  0, 0)				  \
+	 : scm_cell (tc, (scm_t_bits)(data)));				  \
+    if (scm_smobs[_smobnum].free)					  \
+      {									  \
+	GC_finalization_proc _prev_finalizer;				  \
+	GC_PTR _prev_finalizer_data;					  \
+									  \
+	GC_REGISTER_FINALIZER_NO_ORDER (SCM2PTR (z), scm_i_finalize_smob, \
+					NULL,				  \
+					&_prev_finalizer,		  \
+					&_prev_finalizer_data);		  \
+      }									  \
+  }									  \
+while (0)
+
+#define SCM_RETURN_NEWSMOB(tc, data)			\
+  do { SCM __SCM_smob_answer;				\
+       SCM_NEWSMOB (__SCM_smob_answer, (tc), (data));	\
+       return __SCM_smob_answer;			\
   } while (0)
 
-#define SCM_NEWSMOB2(z, tc, data1, data2) \
-do { \
-  z = scm_double_cell ((tc), (scm_t_bits)(data1), (scm_t_bits)(data2), 0); \
-} while (0)
+#define SCM_NEWSMOB2(z, tc, data1, data2)	\
+  SCM_NEWSMOB3 (z, tc, data1, data2, 0)
 
-#define SCM_RETURN_NEWSMOB2(tc, data1, data2) \
-  do { SCM __SCM_smob_answer; \
-       SCM_NEWSMOB2 (__SCM_smob_answer, (tc), (data1), (data2)); \
-       return __SCM_smob_answer; \
+#define SCM_RETURN_NEWSMOB2(tc, data1, data2)				\
+  do { SCM __SCM_smob_answer;						\
+       SCM_NEWSMOB2 (__SCM_smob_answer, (tc), (data1), (data2));	\
+       return __SCM_smob_answer;					\
   } while (0)
 
-#define SCM_NEWSMOB3(z, tc, data1, data2, data3) \
-do { \
-  z = scm_double_cell ((tc), (scm_t_bits)(data1), \
-                       (scm_t_bits)(data2), (scm_t_bits)(data3)); \
-} while (0)
+#define SCM_NEWSMOB3(z, tc, data1, data2, data3)			  \
+do									  \
+  {									  \
+    register scm_t_bits _smobnum = SCM_TC2SMOBNUM (tc);			  \
+    z = (scm_smobs[_smobnum].mark					  \
+	 ? scm_i_new_smob_with_mark_proc (tc, (scm_t_bits)(data1),	  \
+					  (scm_t_bits)(data2),		  \
+					  (scm_t_bits)(data3))		  \
+	 : scm_double_cell ((tc), (scm_t_bits)(data1),			  \
+			    (scm_t_bits)(data2),			  \
+			    (scm_t_bits)(data3)));			  \
+    if (scm_smobs[_smobnum].free)					  \
+      {									  \
+	GC_finalization_proc _prev_finalizer;				  \
+	GC_PTR _prev_finalizer_data;					  \
+									  \
+	GC_REGISTER_FINALIZER_NO_ORDER (SCM2PTR (z), scm_i_finalize_smob, \
+					NULL,				  \
+					&_prev_finalizer,		  \
+					&_prev_finalizer_data);		  \
+      }									  \
+  }									  \
+while (0)
 
-#define SCM_RETURN_NEWSMOB3(tc, data1, data2, data3) \
-  do { SCM __SCM_smob_answer; \
-       SCM_NEWSMOB3 (__SCM_smob_answer, (tc), (data1), (data2), (data3)); \
-       return __SCM_smob_answer; \
+#define SCM_RETURN_NEWSMOB3(tc, data1, data2, data3)			    \
+  do { SCM __SCM_smob_answer;						    \
+       SCM_NEWSMOB3 (__SCM_smob_answer, (tc), (data1), (data2), (data3));   \
+       return __SCM_smob_answer;					    \
   } while (0)
 
 
@@ -120,6 +160,7 @@ SCM_API long scm_numsmob;
 SCM_API scm_smob_descriptor scm_smobs[];
 
 SCM_API void scm_i_set_smob_flags (SCM x, scm_t_bits data);
+SCM_API void scm_i_finalize_smob (GC_PTR obj, GC_PTR data);
 
 
 
