@@ -1,4 +1,4 @@
-/* Copyright (C) 1996,1997,1998,1999,2000,2001, 2003, 2004, 2006, 2007 Free Software Foundation, Inc.
+/* Copyright (C) 1996,1997,1998,1999,2000,2001, 2003, 2004, 2006, 2007, 2008, 2009 Free Software Foundation, Inc.
  * 
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -30,6 +30,7 @@
 #include "libguile/hashtab.h"
 #include "libguile/ports.h"
 #include "libguile/strings.h"
+#include "libguile/srfi-13.h"
 
 #include "libguile/validate.h"
 #include "libguile/struct.h"
@@ -40,6 +41,8 @@
 #include <string.h>
 #endif
 
+#include "libguile/bdw-gc.h"
+
 
 
 static SCM required_vtable_fields = SCM_BOOL_F;
@@ -61,9 +64,9 @@ SCM_DEFINE (scm_make_struct_layout, "make-struct-layout", 1, 0, 0,
 {
   SCM new_sym;
   SCM_VALIDATE_STRING (1, fields);
+  scm_t_wchar c;
 
   { /* scope */
-    const char * field_desc;
     size_t len;
     int x;
 
@@ -72,11 +75,9 @@ SCM_DEFINE (scm_make_struct_layout, "make-struct-layout", 1, 0, 0,
       SCM_MISC_ERROR ("odd length field specification: ~S", 
 		      scm_list_1 (fields));
 
-    field_desc = scm_i_string_chars (fields);
-
     for (x = 0; x < len; x += 2)
       {
-	switch (field_desc[x])
+	switch (c = scm_i_string_ref (fields, x))
 	  {
 	  case 'u':
 	  case 'p':
@@ -88,13 +89,13 @@ SCM_DEFINE (scm_make_struct_layout, "make-struct-layout", 1, 0, 0,
 	    break;
 	  default:
 	    SCM_MISC_ERROR ("unrecognized field type: ~S", 
-			    scm_list_1 (SCM_MAKE_CHAR (field_desc[x])));
+			    scm_list_1 (SCM_MAKE_CHAR (c)));
 	  }
 
-	switch (field_desc[x + 1])
+	switch (c = scm_i_string_ref (fields, x + 1))
 	  {
 	  case 'w':
-	    if (field_desc[x] == 's')
+	    if (scm_i_string_ref (fields, x) == 's')
 	      SCM_MISC_ERROR ("self fields not writable", SCM_EOL);
 	  case 'r':
 	  case 'o':
@@ -102,7 +103,7 @@ SCM_DEFINE (scm_make_struct_layout, "make-struct-layout", 1, 0, 0,
 	  case 'R':
 	  case 'W':
 	  case 'O':
-	    if (field_desc[x] == 's')
+	    if (scm_i_string_ref (fields, x) == 's')
 	      SCM_MISC_ERROR ("self fields not allowed in tail array", 
 			      SCM_EOL);
 	    if (x != len - 2)
@@ -111,12 +112,12 @@ SCM_DEFINE (scm_make_struct_layout, "make-struct-layout", 1, 0, 0,
 	    break;
 	  default:
 	    SCM_MISC_ERROR ("unrecognized ref specification: ~S",
-			    scm_list_1 (SCM_MAKE_CHAR (field_desc[x + 1])));
+			    scm_list_1 (SCM_MAKE_CHAR (c)));
 	  }
 #if 0
-	if (field_desc[x] == 'd')
+	if (scm_i_string_ref (fields, x, 'd'))
 	  {
-	    if (field_desc[x + 2] != '-')
+	    if (!scm_i_string_ref (fields, x+2, '-'))
 	      SCM_MISC_ERROR ("missing dash field at position ~A",
 			      scm_list_1 (scm_from_int (x / 2)));
 	    x += 2;
@@ -138,18 +139,18 @@ SCM_DEFINE (scm_make_struct_layout, "make-struct-layout", 1, 0, 0,
 static void
 scm_struct_init (SCM handle, SCM layout, scm_t_bits * mem, int tail_elts, SCM inits)
 {
-  unsigned const char *fields_desc =
-    (unsigned const char *) scm_i_symbol_chars (layout) - 2;
-  unsigned char prot = 0;
+  scm_t_wchar prot = 0;
   int n_fields = scm_i_symbol_length (layout) / 2;
   int tailp = 0;
+  int i;
 
+  i = -2;
   while (n_fields)
     {
       if (!tailp)
 	{
-	  fields_desc += 2;
-	  prot = fields_desc[1];
+	  i += 2;
+	  prot = scm_i_symbol_ref (layout, i+1);
 	  if (SCM_LAYOUT_TAILP (prot))
 	    {
 	      tailp = 1;
@@ -160,8 +161,7 @@ scm_struct_init (SCM handle, SCM layout, scm_t_bits * mem, int tail_elts, SCM in
 		break;
 	    }
 	}
-      
-      switch (*fields_desc)
+      switch (scm_i_symbol_ref (layout, i))
 	{
 #if 0
 	case 'i':
@@ -237,7 +237,8 @@ SCM_DEFINE (scm_struct_vtable_p, "struct-vtable?", 1, 0, 0,
 {
   SCM layout;
   scm_t_bits * mem;
-  int tmp;
+  SCM tmp;
+  size_t len;
 
   if (!SCM_STRUCTP (x))
     return SCM_BOOL_F;
@@ -248,11 +249,14 @@ SCM_DEFINE (scm_struct_vtable_p, "struct-vtable?", 1, 0, 0,
       < scm_i_string_length (required_vtable_fields))
     return SCM_BOOL_F;
 
-  tmp = strncmp (scm_i_symbol_chars (layout),
-		 scm_i_string_chars (required_vtable_fields),
-		 scm_i_string_length (required_vtable_fields));
-  scm_remember_upto_here_1 (required_vtable_fields);
-  if (tmp)
+  len = scm_i_string_length (required_vtable_fields);
+  tmp = scm_string_eq (scm_symbol_to_string (layout), 
+		       required_vtable_fields, 
+		       scm_from_size_t (0), 
+		       scm_from_size_t (len), 
+		       scm_from_size_t (0),
+		       scm_from_size_t (len));
+  if (scm_is_false (tmp))
     return SCM_BOOL_F;
 
   mem = SCM_STRUCT_DATA (x);
@@ -305,97 +309,53 @@ scm_alloc_struct (int n_words, int n_extra, const char *what)
   p = (scm_t_bits *) (((scm_t_bits) p + 7) & ~7);
 
   /* Initialize a few fields as described above.  */
-  p[scm_struct_i_free] = (scm_t_bits) scm_struct_free_standard;
+  p[scm_struct_i_free] = (scm_t_bits) 0;
   p[scm_struct_i_ptr] = (scm_t_bits) block;
   p[scm_struct_i_n_words] = n_words;
   p[scm_struct_i_flags] = 0;
 
+  /* Since `SCM' objects will record either P or P + SCM_TC3_STRUCT, we need
+     to register them as valid displacements.  Fortunately, only a handful of
+     N_EXTRA values are used in core Guile.  */
+  GC_REGISTER_DISPLACEMENT ((char *)p - (char *)block);
+  GC_REGISTER_DISPLACEMENT ((char *)p - (char *)block + scm_tc3_struct);
+
   return p;
 }
 
-void
-scm_struct_free_0 (scm_t_bits * vtable SCM_UNUSED,
-		   scm_t_bits * data SCM_UNUSED)
+
+/* Finalization.  */
+
+
+/* Invoke the finalizer of the struct pointed to by PTR.  */
+static void
+struct_finalizer_trampoline (GC_PTR ptr, GC_PTR unused_data)
 {
+  SCM obj = PTR2SCM (ptr);
+
+  /* XXX - use less explicit code. */
+  scm_t_bits word0 = SCM_CELL_WORD_0 (obj) - scm_tc3_struct;
+  scm_t_bits *vtable_data = (scm_t_bits *) word0;
+  scm_t_bits *data = SCM_STRUCT_DATA (obj);
+  scm_t_struct_free free_struct_data
+    = ((scm_t_struct_free) vtable_data[scm_struct_i_free]);
+
+  SCM_SET_CELL_TYPE (obj, scm_tc3_struct);
+
+#if 0
+  /* A sanity check.  However, this check can fail if the free function
+     changed between the `make-struct' time and now.  */
+  if (free_struct_data != (scm_t_struct_free)unused_data)
+    abort ();
+#endif
+
+  if (free_struct_data)
+    free_struct_data (vtable_data, data);
 }
 
-void
-scm_struct_free_light (scm_t_bits * vtable, scm_t_bits * data)
-{
-  size_t n = vtable [scm_struct_i_size] & ~SCM_STRUCTF_MASK;
-  scm_gc_free (data, n, "struct");
-}
 
-void
-scm_struct_free_standard (scm_t_bits * vtable SCM_UNUSED, scm_t_bits * data)
-{
-  size_t n = (data[scm_struct_i_n_words] + scm_struct_n_extra_words)
-	     * sizeof (scm_t_bits) + 7;
-  scm_gc_free ((void *) data[scm_struct_i_ptr], n, "heavy struct");
-}
-
-void
-scm_struct_free_entity (scm_t_bits * vtable SCM_UNUSED, scm_t_bits * data)
-{
-  size_t n = (data[scm_struct_i_n_words] + scm_struct_entity_n_extra_words)
-	     * sizeof (scm_t_bits) + 7;
-  scm_gc_free ((void *) data[scm_struct_i_ptr], n, "entity struct");
-}
-
-static void *
-scm_struct_gc_init (void *dummy1 SCM_UNUSED,
-		    void *dummy2 SCM_UNUSED,
-		    void *dummy3 SCM_UNUSED)
-{
-  scm_i_structs_to_free = SCM_EOL;
-  return 0;
-}
-
-static void *
-scm_free_structs (void *dummy1 SCM_UNUSED,
-		  void *dummy2 SCM_UNUSED,
-		  void *dummy3 SCM_UNUSED)
-{
-  SCM newchain = scm_i_structs_to_free;
-  do
-    {
-      /* Mark vtables in GC chain.  GC mark set means delay freeing. */
-      SCM chain = newchain;
-      while (!scm_is_null (chain))
-	{
-	  SCM vtable = SCM_STRUCT_VTABLE (chain);
-	  if (SCM_STRUCT_GC_CHAIN (vtable) != 0 && vtable != chain)
-	    SCM_SET_STRUCT_MARK (vtable);
-	  chain = SCM_STRUCT_GC_CHAIN (chain);
-	}
-      /* Free unmarked structs.  */
-      chain = newchain;
-      newchain = SCM_EOL;
-      while (!scm_is_null (chain))
-	{
-	  SCM obj = chain;
-	  chain = SCM_STRUCT_GC_CHAIN (chain);
-	  if (SCM_STRUCT_MARK_P (obj))
-	    {
-	      SCM_CLEAR_STRUCT_MARK (obj);
-	      SCM_SET_STRUCT_GC_CHAIN (obj, newchain);
-	      newchain = obj;
-	    }
-	  else
-	    {
-	      scm_t_bits * vtable_data = SCM_STRUCT_VTABLE_DATA (obj);
-	      scm_t_bits * data = SCM_STRUCT_DATA (obj);
-	      scm_t_struct_free free_struct_data
-		= ((scm_t_struct_free) vtable_data[scm_struct_i_free]);
-	      SCM_SET_CELL_TYPE (obj, scm_tc_free_cell);
-	      free_struct_data (vtable_data, data);
-	    }
-	}
-    }
-  while (!scm_is_null (newchain));
-  return 0;
-}
 
+
 SCM_DEFINE (scm_make_struct, "make-struct", 2, 0, 1, 
             (SCM vtable, SCM tail_array_size, SCM init),
 	    "Create a new structure.\n\n"
@@ -422,13 +382,15 @@ SCM_DEFINE (scm_make_struct, "make-struct", 2, 0, 1,
   SCM layout;
   size_t basic_size;
   size_t tail_elts;
-  scm_t_bits * data;
+  scm_t_bits *data, *c_vtable;
   SCM handle;
 
   SCM_VALIDATE_VTABLE (1, vtable);
   SCM_VALIDATE_REST_ARGUMENT (init);
 
-  layout = SCM_PACK (SCM_STRUCT_DATA (vtable) [scm_vtable_index_layout]);
+  c_vtable = SCM_STRUCT_DATA (vtable);
+
+  layout = SCM_PACK (c_vtable [scm_vtable_index_layout]);
   basic_size = scm_i_symbol_length (layout) / 2;
   tail_elts = scm_to_size_t (tail_array_size);
 
@@ -458,7 +420,7 @@ SCM_DEFINE (scm_make_struct, "make-struct", 2, 0, 1,
      need for a lock on the section below, as it does not access or update
      any globals, so the critical section has been removed. */
 
-  if (SCM_STRUCT_DATA (vtable)[scm_struct_i_flags] & SCM_STRUCTF_ENTITY)
+  if (c_vtable[scm_struct_i_flags] & SCM_STRUCTF_ENTITY)
     {
       data = scm_alloc_struct (basic_size + tail_elts,
 			       scm_struct_entity_n_extra_words,
@@ -470,10 +432,25 @@ SCM_DEFINE (scm_make_struct, "make-struct", 2, 0, 1,
     data = scm_alloc_struct (basic_size + tail_elts,
 			     scm_struct_n_extra_words,
 			     "struct");
-  handle = scm_double_cell ((((scm_t_bits) SCM_STRUCT_DATA (vtable))
+  handle = scm_double_cell ((((scm_t_bits) c_vtable)
 			     + scm_tc3_struct),
 			    (scm_t_bits) data, 0, 0);
 
+  if (c_vtable[scm_struct_i_free])
+    {
+      /* Register a finalizer for the newly created instance.  */
+      GC_finalization_proc prev_finalizer;
+      GC_PTR prev_finalizer_data;
+      scm_t_struct_free free_struct =
+	(scm_t_struct_free)c_vtable[scm_struct_i_free];
+
+      GC_REGISTER_FINALIZER_NO_ORDER (SCM2PTR (handle),
+				      struct_finalizer_trampoline,
+				      free_struct,
+				      &prev_finalizer,
+				      &prev_finalizer_data);
+    }
+
   scm_struct_init (handle, layout, data, tail_elts, init);
 
   return handle;
@@ -646,8 +623,7 @@ SCM_DEFINE (scm_struct_ref, "struct-ref", 2, 0, 0,
   size_t layout_len;
   size_t p;
   scm_t_bits n_fields;
-  const char *fields_desc;
-  char field_type = 0;
+  scm_t_wchar field_type = 0;
   
 
   SCM_VALIDATE_STRUCT (1, handle);
@@ -656,7 +632,6 @@ SCM_DEFINE (scm_struct_ref, "struct-ref", 2, 0, 0,
   data = SCM_STRUCT_DATA (handle);
   p = scm_to_size_t (pos);
 
-  fields_desc = scm_i_symbol_chars (layout);
   layout_len = scm_i_symbol_length (layout);
   if (SCM_STRUCT_VTABLE_FLAGS (handle) & SCM_STRUCTF_LIGHT)
     /* no extra words */
@@ -668,9 +643,9 @@ SCM_DEFINE (scm_struct_ref, "struct-ref", 2, 0, 0,
 
   if (p * 2 < layout_len)
     {
-      char ref;
-      field_type = fields_desc[p * 2];
-      ref = fields_desc[p * 2 + 1];
+      scm_t_wchar ref;
+      field_type = scm_i_symbol_ref (layout, p * 2);
+      ref = scm_i_symbol_ref (layout, p * 2 + 1);
       if ((ref != 'r') && (ref != 'w'))
 	{
 	  if ((ref == 'R') || (ref == 'W'))
@@ -679,8 +654,8 @@ SCM_DEFINE (scm_struct_ref, "struct-ref", 2, 0, 0,
 	    SCM_MISC_ERROR ("ref denied for field ~A", scm_list_1 (pos));
 	}
     }
-  else if (fields_desc[layout_len - 1] != 'O')    
-    field_type = fields_desc[layout_len - 2];
+  else if (scm_i_symbol_ref (layout, layout_len - 1) != 'O')
+    field_type = scm_i_symbol_ref(layout, layout_len - 2);
   else
     SCM_MISC_ERROR ("ref denied for field ~A", scm_list_1 (pos));
   
@@ -728,8 +703,7 @@ SCM_DEFINE (scm_struct_set_x, "struct-set!", 3, 0, 0,
   size_t layout_len;
   size_t p;
   int n_fields;
-  const char *fields_desc;
-  char field_type = 0;
+  scm_t_wchar field_type = 0;
 
   SCM_VALIDATE_STRUCT (1, handle);
 
@@ -737,7 +711,6 @@ SCM_DEFINE (scm_struct_set_x, "struct-set!", 3, 0, 0,
   data = SCM_STRUCT_DATA (handle);
   p = scm_to_size_t (pos);
 
-  fields_desc = scm_i_symbol_chars (layout);
   layout_len = scm_i_symbol_length (layout);
   if (SCM_STRUCT_VTABLE_FLAGS (handle) & SCM_STRUCTF_LIGHT)
     /* no extra words */
@@ -750,13 +723,13 @@ SCM_DEFINE (scm_struct_set_x, "struct-set!", 3, 0, 0,
   if (p * 2 < layout_len)
     {
       char set_x;
-      field_type = fields_desc[p * 2];
-      set_x = fields_desc [p * 2 + 1];
+      field_type = scm_i_symbol_ref (layout, p * 2);
+      set_x = scm_i_symbol_ref (layout, p * 2 + 1);
       if (set_x != 'w')
 	SCM_MISC_ERROR ("set! denied for field ~A", scm_list_1 (pos));
     }
-  else if (fields_desc[layout_len - 1] == 'W')    
-    field_type = fields_desc[layout_len - 2];
+  else if (scm_i_symbol_ref (layout, layout_len - 1) == 'W')    
+    field_type = scm_i_symbol_ref (layout, layout_len - 2);
   else
     SCM_MISC_ERROR ("set! denied for field ~A", scm_list_1 (pos));
   
@@ -894,11 +867,7 @@ scm_print_struct (SCM exp, SCM port, scm_print_state *pstate)
 void
 scm_struct_prehistory ()
 {
-  scm_i_structs_to_free = SCM_EOL;
-  scm_c_hook_add (&scm_before_sweep_c_hook, scm_struct_gc_init, 0, 0);
-  /* With lazy sweep GC, the point at which the entire heap is swept
-     is just before the mark phase. */
-  scm_c_hook_add (&scm_before_mark_c_hook, scm_free_structs, 0, 0);
+  /* Empty.  */
 }
 
 void
