@@ -63,30 +63,18 @@ typedef struct scm_i_thread {
   int canceled;
   int exited;
 
+  /* Boolean indicating whether the thread is in guile mode.  */
+  int guile_mode;
+
   SCM sleep_object;
   scm_i_pthread_mutex_t *sleep_mutex;
   scm_i_pthread_cond_t sleep_cond;
   int sleep_fd, sleep_pipe[2];
 
-  /* This mutex represents this threads right to access the heap.
-     That right can temporarily be taken away by the GC.
-  */
-  scm_i_pthread_mutex_t heap_mutex;
-
-  /* Boolean tracking whether the above mutex is currently locked by
-     this thread.  This is equivalent to whether or not the thread is
-     in "Guile mode".  This field doesn't need any protection because
-     it is only ever set or tested by the owning thread.
-  */
-  int heap_mutex_locked_by_self;
-
-  /* The freelists of this thread.  Each thread has its own lists so
-     that they can all allocate concurrently.
-  */
-  SCM freelist, freelist2;
-  int clear_freelists_p; /* set if GC was done while thread was asleep */
-  int gc_running_p;      /* non-zero while this thread does GC or a
-			    sweep. */
+  /* Information about the Boehm-GC mark stack during the mark phase.  This
+     is used by `scm_gc_mark ()'.  */
+  void *current_mark_stack_ptr;
+  void *current_mark_stack_limit;
 
   /* Other thread local things.
    */
@@ -123,7 +111,7 @@ typedef struct scm_i_thread {
   SCM vm;
   SCM_STACKITEM *base;
   SCM_STACKITEM *top;
-  jmp_buf regs;
+  scm_i_jmp_buf regs;
 #ifdef __ia64__
   void *register_backing_store_base;
   scm_t_contregs *pending_rbs_continuation;
@@ -151,26 +139,16 @@ SCM_INTERNAL void *scm_i_with_guile_and_parent (void *(*func)(void *),
 						void *data, SCM parent);
 
 
-extern int scm_i_thread_go_to_sleep;
+void scm_threads_prehistory (SCM_STACKITEM *);
+void scm_threads_init_first_thread (void);
 
-SCM_INTERNAL void scm_i_thread_put_to_sleep (void);
-SCM_INTERNAL void scm_i_thread_wake_up (void);
-SCM_INTERNAL void scm_i_thread_invalidate_freelists (void);
-void scm_i_thread_sleep_for_gc (void);
-
-SCM_INTERNAL void scm_threads_prehistory (SCM_STACKITEM *);
-SCM_INTERNAL void scm_threads_init_first_thread (void);
-SCM_INTERNAL void scm_threads_mark_stacks (void);
 SCM_INTERNAL void scm_init_threads (void);
 SCM_INTERNAL void scm_init_thread_procs (void);
 SCM_INTERNAL void scm_init_threads_default_dynamic_state (void);
 
 
 #define SCM_THREAD_SWITCHING_CODE \
-do { \
-  if (scm_i_thread_go_to_sleep) \
-    scm_i_thread_sleep_for_gc (); \
-} while (0)
+  do { } while (0)
 
 SCM_API SCM scm_call_with_new_thread (SCM thunk, SCM handler);
 SCM_API SCM scm_yield (void);
